/***********************************************************************

    This file is part of
    FEINS, Finite Element Incompressible Navier-Stokes solver,
    which is expanding to a more general FEM solver and toolbox,
    Copyright (C) 2003--2013, Rene Schneider 
    <rene.schneider@mathematik.tu-chemnitz.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.

    Minor contributions to this program (for example bug-fixes and
    minor extensions) by third parties automatically transfer the
    copyright to the general author of FEINS, to maintain the
    possibility of commercial re-licensing. If you contribute but wish
    to keep the copyright of your contribution, make that clear in
    your contribution!

    Non-GPL licenses to this program are available upon request from
    the author.

************************************************************************/
#include "feins_macros.h"
#include "datastruc.h"
#include "mesh.h"
#include "assembly.h"
#include "sparse.h"
#include "lin_solver.h"
#include "linsolve_hypre_AMG.h"
#include "gen_aux.h"


//#include <unistd.h>
#include <string.h>

#define TIMEGET {\
   gettimeofday(&tv,&tz); \
   sec=tv.tv_sec; \
   musec=tv.tv_usec; \
   ti=((double)sec+((double)musec)*1e-6); }


int main(int argc, char *argv[])
{
  char solver[50]="no_solver_set";
  struct sparse *Ks;
  struct vector rhs, *x, *old_x;
  struct mesh msh1;
  struct projector1 P;

  struct multilvl ml;
  struct mgdata mg;
  struct coarse_mat *cmat;

  struct solver_settings set;

  int  err, iter, lvl, lmax, level0;
  FIDX i, vx_nr, vx_nr_old;

  double resi, atol, rtol;

  struct timeval tv;
  struct timezone tz;
  int sec, musec;
  double t0, ti, t0l;


  if (argc>1)
    {
      printf("meshfile: %s \n", argv[1]);
    }
  else
    {
      printf("main: no mesh specified!\n");
      return FAIL;
    }
  err=mesh_read_file_t1( &msh1, &set, argv[1] ); /* */
  FUNCTION_FAILURE_HANDLE( err, mesh_read_file_t1, main);

  TIMEGET;
  t0=ti;


  /* set refinement levels */
  level0  = set.refine_ini;
  lmax    = set.refine_steps;

  for (i=0; i<level0 ; i++)
    {
      err=mesh_refine_uniform_t1( &msh1 );
      FUNCTION_FAILURE_HANDLE( err, mesh_refine_uniform_t1, main);
    }

  /* reset hierarchy */
  msh1.hi_nr=0;
  msh1.lvl=-1;
  for (i=0; i<msh1.eg_nr; i++)
    {
      msh1.edge[i*msh1.eg_w+MCT1EGLVL]=-1;
    }


  /* allocate memory for the sparse matrices */
  TRY_MALLOC( Ks,  lmax+1, struct sparse, main);
  old_x=NULL;
  vx_nr_old=0;

  for (lvl=0; lvl<=lmax; lvl ++)
    { 
      if (lvl != 0)
	{
	  err=mesh_refine_uniform_t1( &msh1 );
	  FUNCTION_FAILURE_HANDLE( err, mesh_refine_uniform_t1, main);
	}
      vx_nr=msh1.vx_nr;

      /* allocate memory for x, rhs, K, P */
      TRY_MALLOC( x,  1, struct vector, main);
      err=vector_alloc( x, vx_nr );
      FUNCTION_FAILURE_HANDLE( err, vector_alloc, main);
      
      err=vector_alloc( &rhs, vx_nr );
      FUNCTION_FAILURE_HANDLE( err, vector_alloc, main);

      err=sparse_flex_alloc( &Ks[lvl], vx_nr );
      FUNCTION_FAILURE_HANDLE( err, sparse_flex_alloc, main);

      err=projector1_alloc( &P, vx_nr );
      FUNCTION_FAILURE_HANDLE( err, sparse_flex_alloc, main);
      
      /* init multilvl data */
      err=multilvl_init_tx( &msh1, 1, &ml, 1);
      FUNCTION_FAILURE_HANDLE( err, multilvl_init_tx, main);

      /* on coarse mesh, init x to zero, 
	 on finer meshes interpolate solution from old mesh to new */
      if (lvl==0)
	{
	  for (i=0; i<vx_nr; i++)
	    {
	      (*x).V[i]=0.0;
	    }
	}
      else
	{
	  FIDX node;

	  err=mg_init_tx( NULL, &msh1, &ml, &mg, NULL);
	  FUNCTION_FAILURE_HANDLE( err, mg_init_tx, main);
      
	  /* Remark: vx_nr_old is the number of vertices on the coarser
	     mesh (lvl-1), vx_nr is the number on the finer mesh (lvl) */
	  for (i=0; i<vx_nr_old; i++)
	    {
	      MLVLFINDENTRY(node, i, lvl-1, ml);
	      mg.xl[node]=(*old_x).V[i];
	    }
	  
	  err=mg_interpolate_tx( &mg, lvl-1, lvl, mg.xl);	
	  FUNCTION_FAILURE_HANDLE( err, mg_interpolate_tx, main);  
	      
	  /* now copy back to the solution vector */    
	  for (i=0; i<vx_nr; i++)
	    {
	      MLVLFINDENTRY(node, i, lvl, ml);
	      (*x).V[i]=mg.xl[node];
	    }
	  mg_free(&mg);
	}

      TIMEGET;
      t0l=ti;

      /* assembly of equation system */
      err=assem_poisson_t1( &Ks[lvl], &rhs, x, &P, &msh1 );
      FUNCTION_FAILURE_HANDLE( err, assem_poisson_t1, main);


      /* err=sparse_mat_write_file(&Ks[lvl], "visual/2D_poisson_mat.txt");
      FUNCTION_FAILURE_HANDLE( err, sparse_mat_write_file, main); /* */


      /* define the multigrid data */      
      err=mg_init_tx( Ks, &msh1, &ml, &mg, &P);
      FUNCTION_FAILURE_HANDLE( err, mg_init_tx, main);

      /* use coarse grid solver */
      if (lvl==0)
	{
	  /* generate the coarse grid matrix */
	  FIDX *coarse_bc_nodes;
	  FIDX n_c_bc=0;
	  TRY_MALLOC(cmat, 1,  struct coarse_mat, main);
	  TRY_MALLOC(coarse_bc_nodes, P.len, FIDX, main);
	  for (i=0; i<P.len; i++)
	    {
	      FIDX dof, child;
	      dof=P.V[i];
	      MLVLFINDENTRY(child, dof, 0, ml);
	      if (child>=0)
		{
		  coarse_bc_nodes[n_c_bc]=child-ml.nlevl[0+1];
		  n_c_bc++;
		}
	    }
	  err=coarse_mat_set( &Ks[0], n_c_bc, coarse_bc_nodes,
			      1, cmat );
	  FUNCTION_FAILURE_HANDLE( err, coarse_mat_set, main );
	  free(coarse_bc_nodes);
	}

      /* define solver parameter for the V-cycle */
      mg.vcycles=1;
      mg.smooths=1;
      mg.CGC_scale=1.0;
      mg.cmat=cmat;

      atol = set.solver_atol;
      if (lvl == 0)
	{
	  rtol = set.solver_ini_rtol;
	}
      else
	{
	  rtol = set.solver_ref_rtol;
	} /* */

      /* solve equation system */
      resi=0.0;
      iter=0;

      /* strcpy(solver,"PCG_no");
         err=PCG( 10000, 2, atol, rtol, 1, x, &resi, &iter, sparse_mul_mat_vec, 
	 projector1_no_precon, &Ks[lvl], &rhs, &P );
	 FUNCTION_FAILURE_HANDLE( err, PCG, main); /* */

      strcpy(solver,"PCG_MG");
      err=PCG( 10000, 2, atol, rtol, 1, x, &resi, &iter, sparse_mul_mat_vec, 
	       gen_proj_MG_tx, &Ks[lvl], &rhs, &mg );
      FUNCTION_FAILURE_HANDLE( err, PCG, main); /* */

      /*strcpy(solver,"PCG_AMG");
	{
	void *AMGdata;
	err=gen_hypre_AMG_setup(&Ks[lvl], P.len, P.V, 1, &AMGdata, 1 );
	FUNCTION_FAILURE_HANDLE( err, gen_hypre_AMG_setup, main); 
	err=PCG( 10000, 2, atol, rtol, 1, x, &resi, &iter, sparse_mul_mat_vec, 
	gen_hypre_AMG_preco, &Ks[lvl], &rhs, AMGdata );
	FUNCTION_FAILURE_HANDLE( err, PCG, main);
	err=gen_hypre_AMG_free(&AMGdata);
	FUNCTION_FAILURE_HANDLE( err, gen_hypre_AMG_setup, main); 	
	} /* */

      
      /* strcpy(solver,"PCR_no");
	 err=PCR( 10000, 2, atol, rtol, 1, x, &resi, &iter, sparse_mul_mat_vec, 
	 projector1_no_precon, &Ks[lvl], &rhs, &P );
	 FUNCTION_FAILURE_HANDLE( err, PCR, main); /* */


      TIMEGET;

      printf("%s: %4d iterations, |res|=%8.2e, vx_nr= %9"dFIDX",  "
	     "time_level=%10.3e, time_total=%10.3e\n",
	     solver, iter, resi, vx_nr, ti-t0l, ti-t0); /* */


      if (lvl==lmax)
	{
	  err= mesh_write_solution_vtk_t1(&msh1, NULL, 0, x, 1,
					  15,"visual/poisson" );
	  FUNCTION_FAILURE_HANDLE( err, mesh_write_solution_vtk_t1, main);/* */
	}

      mg.cmat=NULL; /* prevent it from beeing destroyed, so we don't
		       need to recompute it every time */
      mg_free(&mg);
      multilvl_free(&ml);


      projector1_free(&P);
      vector_free(&rhs);
      /* keep current solution vector as old solution */
      if (old_x!=NULL)
	{
	  vector_free(old_x);
	  free(old_x);
	}
      old_x=x;
      vx_nr_old=vx_nr;
      x=NULL;
    } /* for lvl */


  /* free remaining memory */
  vector_free(old_x);
  free(old_x);
  
  coarse_mat_free(cmat);
  free(cmat);

  for (lvl=lmax; lvl>=0; lvl--)
    sparse_free(&Ks[lvl]);
  free(Ks);

  mesh_free(&msh1);

  return 1;
}
